#include "module.h"
#include "droptarget.h"

#ifndef NO_SCRIPT_ACTIONS
static void UpdateHover(WndInfo *w, bool bOnMouseMove)
{
  bool bHoveringNow,bMouseOverNow;
  POINT pt;
  RECT r;

  ASSERT_WRITEABLE(w);

  GetCursorPos(&pt);
  GetWindowRect(w->hWnd,&r);

  bMouseOverNow = PtInRect(&r,pt)!=0;
  bHoveringNow = bMouseOverNow && pt.x==w->HoverPoint.x && pt.y==w->HoverPoint.y;

  if (bMouseOverNow && WindowFromPoint(pt)!=w->hWnd)
    bMouseOverNow = false;

  if (bOnMouseMove)
  {
    if (bMouseOverNow && !w->bMouseOver)
    {
      w->bMouseOver = true;
      ExecAction(w,OnMouseOver);
      SetCapture(w->hWnd);
    }
    if (w->bHovering && !bHoveringNow)
      { w->bHovering = false; ExecAction(w,OnUnhover); }
    w->HoverPoint = pt;
    w->HoverTimerID = 
      SetTimer(w->hWnd,HoverTimerEvent,GetHoverTime(w),NULL);
  }
  else
  {
    if (bHoveringNow && !w->bHovering)
      { w->bHovering = true; ExecAction(w,OnHover); }
  }
  if (w->bMouseOver && !bMouseOverNow)
  {
    w->bMouseOver = false; ExecAction(w,OnMouseOut);
    KillTimer(w->hWnd,w->HoverTimerID);
    if (!w->bLButtonDown && !w->bRButtonDown && !w->bMButtonDown)
      ReleaseCapture();
  }
}

void MouseButton(WndInfo *w, unsigned which, bool state)
{
  static bool WndInfo::* const Flag[] =
  {
    &WndInfo::bLButtonDown,
    &WndInfo::bRButtonDown,
    &WndInfo::bMButtonDown,
  };
  const static int EventNo[] =
  {
    OnLButtonDown, OnLButtonUp, OnLButtonClick,
    OnRButtonDown, OnRButtonUp, OnRButtonClick,
    OnMButtonDown, OnMButtonUp, OnMButtonClick,
  };
  bool bMouseOverNow;
  POINT pt;
  RECT r;

  ASSERT_READABLE(w);
  ASSERT(which<3);
  ASSERT((sizeof(EventNo)/sizeof(int))==3*3);
  ASSERT((sizeof(Flag)/sizeof(bool WndInfo::*))==3);

  GetCursorPos(&pt);
  GetWindowRect(w->hWnd,&r);

  bMouseOverNow = PtInRect(&r,pt)!=0;

  switch (state)
  {
  case true: /* button down */ 
    ExecAction(w,EventNo[which*3]);
    break;

  case false: /* button up */ 
    if (w->*Flag[which])
    {
      ExecAction(w,EventNo[which*3+1]);
      if (bMouseOverNow)
        ExecAction(w,EventNo[which*3+2]);
    }
    if (!bMouseOverNow)
      ReleaseCapture();
  }

  w->*Flag[which] = state;
}
#endif /* NO_SCRIPT_ACTIONS */ 


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	WndInfo *w = (WndInfo *)GetWindowLong(hwnd,UGWL_WNDINFO_PTR);
	switch (message) {
	case WM_SHOWWINDOW:
		{
			HDC hdcScreen = GetDC(NULL);

			if (w->hbmBackBuffer==NULL)
				w->hbmBackBuffer = CreateCompatibleBitmap(hdcScreen,w->Pos.Width.Value,w->Pos.Height.Value);

			ReleaseDC(NULL,hdcScreen);

			RedrawWnd(w);
		}
		break;

	case WM_DESTROY:
		{
     #ifndef NO_SCRIPT_ACTIONS
      /* OnCreate is in CreateWnd() */ 
      ExecAction(w,OnDestroy);
     #endif

      w->DropTarget->Revoke();
      delete w->DropTarget;
			if (w->hbmBackBuffer!=NULL) DeleteObject(w->hbmBackBuffer);
		}
		break;

	case WM_PAINT:
		{
      /* not used for layered windows */ 

			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hwnd,&ps);
			HDC hdcWnd = CreateCompatibleDC(hdc);
			HBITMAP hbmOld = (HBITMAP)SelectObject(hdcWnd,w->hbmBackBuffer);

			BitBlt(hdc,0,0,w->Pos.Width.Value,w->Pos.Height.Value,hdcWnd,0,0,SRCCOPY);

			SelectObject(hdcWnd,hbmOld);
			DeleteDC(hdcWnd);

			EndPaint(hwnd,&ps);
		}
		break;

#ifndef NO_SCRIPT_ACTIONS
  case WM_LBUTTONDOWN: MouseButton(w, 0, true ); break;
  case WM_LBUTTONUP:   MouseButton(w, 0, false); break;
  case WM_RBUTTONDOWN: MouseButton(w, 1, true ); break;
  case WM_RBUTTONUP:   MouseButton(w, 1, false); break;
  case WM_MBUTTONDOWN: MouseButton(w, 2, true ); break;
  case WM_MBUTTONUP:   MouseButton(w, 2, false); break;
  case WM_MOUSEMOVE:   UpdateHover(w,true);
                       ExecAction(w,OnMouseMove  ); break;

  case WM_TIMER:
    if (wParam == HoverTimerEvent)
  case WM_CAPTURECHANGED:
      UpdateHover(w,false);
    break;
#endif

  case WM_SYSCOLORCHANGE:
    if (CalcColorSet(&w->cSkin))
      RedrawWnd(w);
    break;

#ifdef LM_MONITORSREFRESHED
  case LM_MONITORSREFRESHED:
    LSFreeDesktopCache();
    UpdateWndPos(w);
    break;
#endif

	case LM_GETREVID:
		{
  		PTSTR buffer = (PTSTR)lParam;

			if (wParam == 0)
			{
				lstrcpy(buffer, ModuleName);
				lstrcat(buffer, ".dll: ");
				lstrcat(buffer, ModuleVersion);
				lstrcat(buffer, " (");
				lstrcat(buffer, ModuleAuthor);
				lstrcat(buffer, ")");
			}
			else
			{
				*buffer = 0;
			}
			return lstrlen(buffer);
		}

	default:
		return DefWindowProc(hwnd,message,wParam,lParam);
	}

	return 0;
}